@@ -1,8 +1,8 @@
 from gettext import gettext as _
-import os
 import shutil
 import tempfile
 
+from django.core.files.storage import default_storage as storage
 from packaging.requirements import Requirement
 from rest_framework import serializers
 
@@ -12,6 +12,7 @@ from pulpcore.plugin import serializers as core_serializers
 from pulp_python.app import models as python_models
 from pulp_python.app.tasks.upload import DIST_EXTENSIONS, DIST_TYPES
 from pulp_python.app.utils import parse_project_metadata
+from pulp_python.app.urls import PYPI_API_HOSTNAME
 
 
 class PythonRepositorySerializer(core_serializers.RepositorySerializer):
@@ -19,19 +20,68 @@ class PythonRepositorySerializer(core_serializers.RepositorySerializer):
     Serializer for Python Repositories.
     """
 
+    autopublish = serializers.BooleanField(
+        help_text=_(
+            "Whether to automatically create publications for new repository versions, "
+            "and update any distributions pointing to this repository."
+        ),
+        default=False,
+        required=False,
+    )
+
     class Meta:
-        fields = core_serializers.RepositorySerializer.Meta.fields
+        fields = core_serializers.RepositorySerializer.Meta.fields + ("autopublish",)
         model = python_models.PythonRepository
 
 
-class PythonDistributionSerializer(core_serializers.PublicationDistributionSerializer):
+class PythonDistributionSerializer(core_serializers.DistributionSerializer):
     """
     Serializer for Pulp distributions for the Python type.
-
     """
 
+    publication = core_serializers.DetailRelatedField(
+        required=False,
+        help_text=_("Publication to be served"),
+        view_name_pattern=r"publications(-.*/.*)?-detail",
+        queryset=core_models.Publication.objects.exclude(complete=False),
+        allow_null=True,
+    )
+    base_url = serializers.SerializerMethodField(read_only=True)
+    allow_uploads = serializers.BooleanField(
+        default=True,
+        help_text=_("Allow packages to be uploaded to this index.")
+    )
+
+    def get_base_url(self, obj):
+        """Gets the base url."""
+        return f"{PYPI_API_HOSTNAME}/pypi/{obj.base_path}/"
+
+    def validate(self, data):
+        """
+        Ensure publication and repository are not set at the same time.
+
+        This is needed here till https://pulp.plan.io/issues/8761 is resolved.
+        """
+        data = super().validate(data)
+        repository_provided = data.get("repository", None)
+        publication_provided = data.get("publication", None)
+
+        if repository_provided and publication_provided:
+            raise serializers.ValidationError(
+                _(
+                    "Only one of the attributes 'repository' and 'publication' "
+                    "may be used simultaneously."
+                )
+            )
+        if repository_provided or publication_provided:
+            data["repository"] = repository_provided
+            data["publication"] = publication_provided
+        return data
+
     class Meta:
-        fields = core_serializers.PublicationDistributionSerializer.Meta.fields
+        fields = core_serializers.DistributionSerializer.Meta.fields + (
+            'publication', "allow_uploads"
+        )
         model = python_models.PythonDistribution
 
 
@@ -59,6 +109,10 @@ class PythonPackageContentSerializer(core_serializers.SingleArtifactContentUploa
         help_text=_('The packages version number.'),
         read_only=True,
     )
+    sha256 = serializers.CharField(
+        default='',
+        help_text=_('The SHA256 digest of this package.'),
+    )
     metadata_version = serializers.CharField(
         help_text=_('Version of the file format'),
         read_only=True,
@@ -71,6 +125,11 @@ class PythonPackageContentSerializer(core_serializers.SingleArtifactContentUploa
         required=False, allow_blank=True,
         help_text=_('A longer description of the package that can run to several paragraphs.')
     )
+    description_content_type = serializers.CharField(
+        required=False, allow_blank=True,
+        help_text=_('A string stating the markup syntax (if any) used in the distribution’s'
+                    ' description, so that tools can intelligently render the description.')
+    )
     keywords = serializers.CharField(
         required=False, allow_blank=True,
         help_text=_('Additional keywords to be used to assist searching for the '
@@ -115,6 +174,10 @@ class PythonPackageContentSerializer(core_serializers.SingleArtifactContentUploa
         required=False, allow_blank=True,
         help_text=_('A browsable URL for the project and a label for it, separated by a comma.')
     )
+    project_urls = serializers.JSONField(
+        required=False, default=dict,
+        help_text=_('A dictionary of labels and URLs for the project.')
+    )
     platform = serializers.CharField(
         required=False, allow_blank=True,
         help_text=_('A comma-separated list of platform specifications, '
@@ -168,22 +231,18 @@ class PythonPackageContentSerializer(core_serializers.SingleArtifactContentUploa
         except KeyError:
             raise serializers.ValidationError(detail={"relative_path": _('This field is required')})
 
-        if python_models.PythonPackageContent.objects.filter(filename=filename):
-            raise serializers.ValidationError(detail={
-                "relative_path": _('This field must be unique')
-            })
-
         # iterate through extensions since splitext does not support things like .tar.gz
         for ext, packagetype in DIST_EXTENSIONS.items():
             if filename.endswith(ext):
                 # Copy file to a temp directory under the user provided filename, we do this
                 # because pkginfo validates that the filename has a valid extension before
                 # reading it
-                with tempfile.TemporaryDirectory() as td:
-                    temp_path = os.path.join(td, filename)
+                with tempfile.NamedTemporaryFile('wb', suffix=filename) as temp_file:
                     artifact = data["artifact"]
-                    shutil.copy2(artifact.file.path, temp_path)
-                    metadata = DIST_TYPES[packagetype](temp_path)
+                    artifact_file = storage.open(artifact.file.name)
+                    shutil.copyfileobj(artifact_file, temp_file)
+                    temp_file.flush()
+                    metadata = DIST_TYPES[packagetype](temp_file.name)
                     metadata.packagetype = packagetype
                     break
         else:
@@ -191,36 +250,33 @@ class PythonPackageContentSerializer(core_serializers.SingleArtifactContentUploa
                 "Extension on {} is not a valid python extension "
                 "(.whl, .exe, .egg, .tar.gz, .tar.bz2, .zip)").format(filename)
             )
+        if data.get("sha256") and data["sha256"] != artifact.sha256:
+            raise serializers.ValidationError(
+                detail={"sha256": _(
+                    "The uploaded artifact's sha256 checksum does not match the one provided"
+                )}
+            )
+        sha256 = artifact.sha256
+        if sha256 and python_models.PythonPackageContent.objects.filter(sha256=sha256).exists():
+            raise serializers.ValidationError(detail={"sha256": _('This field must be unique')})
+
         _data = parse_project_metadata(vars(metadata))
         _data['packagetype'] = metadata.packagetype
         _data['version'] = metadata.version
         _data['filename'] = filename
+        _data['sha256'] = sha256
 
         data.update(_data)
 
-        new_content = python_models.PythonPackageContent.objects.filter(
-            filename=data['filename'],
-            packagetype=data['packagetype'],
-            name=data['name'],
-            version=data['version']
-        )
-
-        if new_content.exists():
-            raise serializers.ValidationError(
-                _(
-                    "There is already a python package with relative path '{path}'."
-                ).format(path=data["relative_path"])
-            )
-
         return data
 
     class Meta:
         fields = core_serializers.SingleArtifactContentUploadSerializer.Meta.fields + (
-            'filename', 'packagetype', 'name', 'version', 'metadata_version', 'summary',
-            'description', 'keywords', 'home_page', 'download_url', 'author', 'author_email',
-            'maintainer', 'maintainer_email', 'license', 'requires_python', 'project_url',
-            'platform', 'supported_platform', 'requires_dist', 'provides_dist',
-            'obsoletes_dist', 'requires_external', 'classifiers'
+            'filename', 'packagetype', 'name', 'version', 'sha256', 'metadata_version', 'summary',
+            'description', 'description_content_type', 'keywords', 'home_page', 'download_url',
+            'author', 'author_email', 'maintainer', 'maintainer_email', 'license',
+            'requires_python', 'project_url', 'project_urls', 'platform', 'supported_platform',
+            'requires_dist', 'provides_dist', 'obsoletes_dist', 'requires_external', 'classifiers'
         )
         model = python_models.PythonPackageContent
 
@@ -232,11 +288,21 @@ class MinimalPythonPackageContentSerializer(PythonPackageContentSerializer):
 
     class Meta:
         fields = core_serializers.SingleArtifactContentUploadSerializer.Meta.fields + (
-            'filename', 'packagetype', 'name', 'version',
+            'filename', 'packagetype', 'name', 'version', 'sha256',
         )
         model = python_models.PythonPackageContent
 
 
+class MultipleChoiceArrayField(serializers.MultipleChoiceField):
+    """
+    A wrapper to make sure this DRF serializer works properly with ArrayFields.
+    """
+
+    def to_internal_value(self, data):
+        """Converts set to list."""
+        return list(super().to_internal_value(data))
+
+
 class PythonRemoteSerializer(core_serializers.RemoteSerializer):
     """
     A Serializer for PythonRemote.
@@ -263,10 +329,30 @@ class PythonRemoteSerializer(core_serializers.RemoteSerializer):
     )
     policy = serializers.ChoiceField(
         help_text=_("The policy to use when downloading content. The possible values include: "
-                    "'immediate', 'on_demand', and 'cache_only'. 'immediate' is the default."),
+                    "'immediate', 'on_demand', and 'streamed'. 'on_demand' is the default."),
         choices=core_models.Remote.POLICY_CHOICES,
         default=core_models.Remote.ON_DEMAND
     )
+    package_types = MultipleChoiceArrayField(
+        required=False,
+        help_text=_("The package types to sync for Python content. Leave blank to get every"
+                    "package type."),
+        choices=python_models.PACKAGE_TYPES,
+        default=list
+    )
+    keep_latest_packages = serializers.IntegerField(
+        required=False,
+        help_text=_("The amount of latest versions of a package to keep on sync, includes"
+                    "pre-releases if synced. Default 0 keeps all versions."),
+        default=0
+    )
+    exclude_platforms = MultipleChoiceArrayField(
+        required=False,
+        help_text=_("List of platforms to exclude syncing Python packages for. Possible values"
+                    "include: windows, macos, freebsd, and linux."),
+        choices=python_models.PLATFORMS,
+        default=list
+    )
 
     def validate_includes(self, value):
         """Validates the includes"""
@@ -292,7 +378,8 @@ class PythonRemoteSerializer(core_serializers.RemoteSerializer):
 
     class Meta:
         fields = core_serializers.RemoteSerializer.Meta.fields + (
-            "includes", "excludes", "prereleases"
+            "includes", "excludes", "prereleases", "package_types", "keep_latest_packages",
+            "exclude_platforms",
         )
         model = python_models.PythonRemote
 
@@ -314,7 +401,7 @@ class PythonBanderRemoteSerializer(serializers.Serializer):
 
     policy = serializers.ChoiceField(
         help_text=_("The policy to use when downloading content. The possible values include: "
-                    "'immediate', 'on_demand', and 'cache_only'. 'immediate' is the default."),
+                    "'immediate', 'on_demand', and 'streamed'. 'on_demand' is the default."),
         choices=core_models.Remote.POLICY_CHOICES,
         default=core_models.Remote.ON_DEMAND
     )
@@ -328,8 +415,8 @@ class PythonPublicationSerializer(core_serializers.PublicationSerializer):
     distributions = core_serializers.DetailRelatedField(
         help_text=_('This publication is currently being hosted as configured by these '
                     'distributions.'),
-        source="python_pythondistribution",
-        view_name="filedistributions-detail",
+        source="distribution_set",
+        view_name="pythondistributions-detail",
         many=True,
         read_only=True,
     )
